@@ -6,13 +6,17 @@
 
 import math
 import os
+import typing
+from io import BytesIO, IOBase
+from mmap import mmap
 from numbers import Number
+from pathlib import Path
 from struct import calcsize, pack, unpack
 
 from ..constants import INT32_T_MAX, INT32_T_MIN, INT64_T_MAX, INT64_T_MIN
 from ..exceptions import CountMinSketchError, InitializationError, NotSupportedError
-from ..hashes import default_fnv_1a
-from ..utilities import is_valid_file
+from ..hashes import HashFuncT, HashResultsT, KeyT, default_fnv_1a
+from ..utilities import MMap, is_valid_file
 
 
 class CountMinSketch(object):
@@ -55,14 +59,14 @@ class CountMinSketch(object):
 
     def __init__(
         self,
-        width=None,
-        depth=None,
-        confidence=None,
-        error_rate=None,
-        filepath=None,
-        hash_function=None,
-    ):
-        """ default initilization function """
+        width: typing.Optional[int] = None,
+        depth: typing.Optional[int] = None,
+        confidence: typing.Optional[float] = None,
+        error_rate: typing.Optional[float] = None,
+        filepath: typing.Optional[typing.Union[str, Path]] = None,
+        hash_function: typing.Optional[HashFuncT] = None,
+    ) -> None:
+        """default initilization function"""
         # default values
         self.__width = 0
         self.__depth = 0
@@ -109,10 +113,10 @@ class CountMinSketch(object):
         if hash_function is None:
             self._hash_function = default_fnv_1a
         else:
-            self._hash_function = hash_function
+            self._hash_function = hash_function  # type: ignore
 
-    def __str__(self):
-        """ string representation of the count min sketch """
+    def __str__(self) -> str:
+        """string representation of the count min sketch"""
         msg = (
             "Count-Min Sketch:\n"
             "\tWidth: {0}\n"
@@ -129,12 +133,18 @@ class CountMinSketch(object):
             self.elements_added,
         )
 
-    def __contains__(self, key):
-        """ setup the `in` keyword """
+    def __contains__(self, key: KeyT) -> bool:
+        """setup the `in` keyword"""
         return True if self.check(key) != 0 else False
 
+    def __bytes__(self) -> bytes:
+        """Export countmin-sketch to `bytes`"""
+        with BytesIO() as f:
+            self.export(f)
+            return f.getvalue()
+
     @property
-    def width(self):
+    def width(self) -> int:
         """int: The width of the count-min sketch
 
         Note:
@@ -142,7 +152,7 @@ class CountMinSketch(object):
         return self.__width
 
     @property
-    def depth(self):
+    def depth(self) -> int:
         """int: The depth of the count-min sketch
 
         Note:
@@ -150,7 +160,7 @@ class CountMinSketch(object):
         return self.__depth
 
     @property
-    def confidence(self):
+    def confidence(self) -> float:
         """float: The confidence of the count-min sketch
 
         Note:
@@ -158,7 +168,7 @@ class CountMinSketch(object):
         return self.__confidence
 
     @property
-    def error_rate(self):
+    def error_rate(self) -> float:
         """float: The error rate of the count-min sketch
 
         Note:
@@ -166,7 +176,7 @@ class CountMinSketch(object):
         return self.__error_rate
 
     @property
-    def elements_added(self):
+    def elements_added(self) -> int:
         """int: The number of elements added to the count-min sketch
 
         Note:
@@ -174,7 +184,7 @@ class CountMinSketch(object):
         return self.__elements_added
 
     @property
-    def query_type(self):
+    def query_type(self) -> str:
         """str: The name of the query type being used
 
         Note:
@@ -189,7 +199,7 @@ class CountMinSketch(object):
         return "min"
 
     @query_type.setter
-    def query_type(self, val):
+    def query_type(self, val: str):
         """set to min query Options='min', 'mean', 'mean-min'
         other values are set to min
         setting to mean is converting to a Count-Mean Sketch
@@ -205,13 +215,13 @@ class CountMinSketch(object):
         else:
             self.__query_method = self.__min_query
 
-    def clear(self):
-        """ Reset the count-min sketch to an empty state """
+    def clear(self) -> None:
+        """Reset the count-min sketch to an empty state"""
         self.__elements_added = 0
         for i, _ in enumerate(self._bins):
             self._bins[i] = 0
 
-    def hashes(self, key, depth=None):
+    def hashes(self, key: KeyT, depth: typing.Optional[int] = None) -> HashResultsT:
         """ Return the hashes based on the provided key
 
             Args:
@@ -224,7 +234,7 @@ class CountMinSketch(object):
         t_depth = self.depth if depth is None else depth
         return self._hash_function(key, t_depth)
 
-    def add(self, key, num_els=1):
+    def add(self, key: KeyT, num_els: int = 1) -> int:
         """ Insert the element `key` into the count-min sketch
 
             Args:
@@ -236,7 +246,7 @@ class CountMinSketch(object):
         hashes = self.hashes(key)
         return self.add_alt(hashes, num_els)
 
-    def add_alt(self, hashes, num_els=1):
+    def add_alt(self, hashes: HashResultsT, num_els: int = 1) -> int:
         """ Insert an element by using the hash representation
 
             Args:
@@ -258,7 +268,7 @@ class CountMinSketch(object):
             self.__elements_added = INT64_T_MAX
         return self.__query_method(sorted(res))
 
-    def remove(self, key, num_els=1):
+    def remove(self, key: KeyT, num_els: int = 1) -> int:
         """ Remove element 'key' from the count-min sketch
 
             Args:
@@ -270,7 +280,7 @@ class CountMinSketch(object):
         hashes = self.hashes(key)
         return self.remove_alt(hashes, num_els)
 
-    def remove_alt(self, hashes, num_els=1):
+    def remove_alt(self, hashes: HashResultsT, num_els: int = 1) -> int:
         """ Remove an element by using the hash representation
 
             Args:
@@ -292,7 +302,7 @@ class CountMinSketch(object):
 
         return self.__query_method(sorted(res))
 
-    def check(self, key):
+    def check(self, key: KeyT) -> int:
         """Check number of times element 'key' is in the count-min sketch
 
         Args:
@@ -302,7 +312,7 @@ class CountMinSketch(object):
         hashes = self.hashes(key)
         return self.check_alt(hashes)
 
-    def check_alt(self, hashes):
+    def check_alt(self, hashes: HashResultsT) -> int:
         """ Check the count-min sketch for an element by using the hash \
             representation
 
@@ -313,19 +323,22 @@ class CountMinSketch(object):
         bins = self.__get_values_sorted(hashes)
         return self.__query_method(bins)
 
-    def export(self, filepath):
+    def export(self, file: typing.Union[Path, str, IOBase, mmap]) -> None:
         """ Export the count-min sketch to disk
 
             Args:
                 filename (str): The filename to which the count-min sketch \
                 will be written. """
-        with open(filepath, "wb") as filepointer:
+        if not isinstance(file, (IOBase, mmap)):
+            with open(file, "wb") as filepointer:
+                self.export(filepointer)  # type: ignore
+        else:
             # write out the bins
             rep = "i" * len(self._bins)
-            filepointer.write(pack(rep, *self._bins))
-            filepointer.write(pack("IIq", self.width, self.depth, self.elements_added))
+            file.write(pack(rep, *self._bins))
+            file.write(pack("IIq", self.width, self.depth, self.elements_added))
 
-    def join(self, second):
+    def join(self, second: "CountMinSketch") -> None:
         """ Join two count-min sketchs into a single count-min sketch; the
             calling count-min sketch will have the resulting combined data
 
@@ -368,26 +381,30 @@ class CountMinSketch(object):
         elif self.elements_added < INT64_T_MIN:
             self.__elements_added = INT64_T_MIN
 
-    def __load(self, filepath):
-        """ load the count-min sketch from file """
-        with open(filepath, "rb") as filepointer:
+    def __load(self, file: typing.Union[Path, str, IOBase, mmap]):
+        """load the count-min sketch from file"""
+        if not isinstance(file, (IOBase, mmap)):
+            file = Path(file)
+            with MMap(file) as filepointer:
+                self.__load(filepointer)
+        else:
             offset = calcsize("IIq")
-            filepointer.seek(offset * -1, os.SEEK_END)
-            mybytes = unpack("IIq", filepointer.read(offset))
+            file.seek(offset * -1, os.SEEK_END)
+            mybytes = unpack("IIq", file.read(offset))
             self.__width = mybytes[0]
             self.__depth = mybytes[1]
             self.__elements_added = mybytes[2]
             self.__confidence = 1 - (1 / math.pow(2, self.depth))
             self.__error_rate = 2 / self.width
 
-            filepointer.seek(0, os.SEEK_SET)
+            file.seek(0, os.SEEK_SET)
             length = self.width * self.depth
             rep = "i" * length
             offset = calcsize(rep)
-            self._bins = list(unpack(rep, filepointer.read(offset)))
+            self._bins = list(unpack(rep, file.read(offset)))
 
-    def __get_values_sorted(self, hashes):
-        """ get the values sorted """
+    def __get_values_sorted(self, hashes: HashResultsT) -> HashResultsT:
+        """get the values sorted"""
         bins = list()
         for i, val in enumerate(hashes):
             t_bin = (val % self.width) + (i * self.width)
@@ -396,16 +413,16 @@ class CountMinSketch(object):
         return bins
 
     @staticmethod
-    def __min_query(results):
-        """ generate the min query; assumes sorted list """
+    def __min_query(results: HashResultsT) -> int:
+        """generate the min query; assumes sorted list"""
         return results[0]
 
-    def __mean_query(self, results):
-        """ generate the mean query; assumes sorted list """
+    def __mean_query(self, results: HashResultsT) -> int:
+        """generate the mean query; assumes sorted list"""
         return sum(results) // self.depth
 
-    def __mean_min_query(self, results):
-        """ generate the mean-min query; assumes sorted list """
+    def __mean_min_query(self, results: HashResultsT) -> int:
+        """generate the mean-min query; assumes sorted list"""
         if results[0] == 0 and results[-1] == 0:
             return 0
         meanmin = list()
@@ -452,13 +469,13 @@ class CountMeanSketch(CountMinSketch):
 
     def __init__(
         self,
-        width=None,
-        depth=None,
-        confidence=None,
-        error_rate=None,
-        filepath=None,
-        hash_function=None,
-    ):
+        width: typing.Optional[int] = None,
+        depth: typing.Optional[int] = None,
+        confidence: typing.Optional[float] = None,
+        error_rate: typing.Optional[float] = None,
+        filepath: str = None,
+        hash_function: typing.Optional[HashFuncT] = None,
+    ) -> None:
         super(CountMeanSketch, self).__init__(width, depth, confidence, error_rate, filepath, hash_function)
         self.query_type = "mean"
 
@@ -493,13 +510,13 @@ class CountMeanMinSketch(CountMinSketch):
 
     def __init__(
         self,
-        width=None,
-        depth=None,
-        confidence=None,
-        error_rate=None,
-        filepath=None,
-        hash_function=None,
-    ):
+        width: typing.Optional[int] = None,
+        depth: typing.Optional[int] = None,
+        confidence: typing.Optional[float] = None,
+        error_rate: typing.Optional[float] = None,
+        filepath: typing.Optional[str] = None,
+        hash_function: typing.Optional[HashFuncT] = None,
+    ) -> None:
         super(CountMeanMinSketch, self).__init__(width, depth, confidence, error_rate, filepath, hash_function)
         self.query_type = "mean-min"
 
@@ -534,29 +551,29 @@ class HeavyHitters(CountMinSketch):
 
     def __init__(
         self,
-        num_hitters=100,
-        width=None,
-        depth=None,
-        confidence=None,
-        error_rate=None,
-        filepath=None,
-        hash_function=None,
-    ):
+        num_hitters: int = 100,
+        width: typing.Optional[int] = None,
+        depth: typing.Optional[int] = None,
+        confidence: typing.Optional[float] = None,
+        error_rate: typing.Optional[float] = None,
+        filepath: typing.Optional[typing.Union[str, Path]] = None,
+        hash_function: typing.Optional[HashFuncT] = None,
+    ) -> None:
 
         super(HeavyHitters, self).__init__(width, depth, confidence, error_rate, filepath, hash_function)
-        self.__top_x = dict()  # top x heavy hitters
+        self.__top_x = dict()  # type: ignore
         self.__top_x_size = 0
         self.__num_hitters = num_hitters
         self.__smallest = 0
 
-    def __str__(self):
-        """ heavy hitters string rep """
+    def __str__(self) -> str:
+        """heavy hitters string rep"""
         msg = super(HeavyHitters, self).__str__()
-        tmp = "Heavy Hitters {0}\n" "\tNumber Hitters: {1}\n" "\tNumber Recorded: {2}"
+        tmp = "Heavy Hitters {0}\n\tNumber Hitters: {1}\n\tNumber Recorded: {2}"
         return tmp.format(msg, self.number_heavy_hitters, self.__top_x_size)
 
     @property
-    def heavy_hitters(self):
+    def heavy_hitters(self) -> typing.Dict[str, int]:
         """dict: Return the heavy hitters, or most common elements
 
         Note:
@@ -564,14 +581,14 @@ class HeavyHitters(CountMinSketch):
         return self.__top_x
 
     @property
-    def number_heavy_hitters(self):
+    def number_heavy_hitters(self) -> int:
         """int: Return the maximum number of heavy hitters being tracked
 
         Note:
             Not settable"""
         return self.__num_hitters
 
-    def add(self, key, num_els=1):
+    def add(self, key: str, num_els: int = 1) -> int:  # type: ignore
         """Add element to heavy hitters
 
         Args:
@@ -584,7 +601,7 @@ class HeavyHitters(CountMinSketch):
         hashes = self.hashes(key)
         return self.add_alt(key, hashes, num_els)
 
-    def add_alt(self, key, hashes, num_els=1):
+    def add_alt(self, key: str, hashes: HashResultsT, num_els: int = 1) -> int:  # type: ignore
         """ Add the element `key` represented as hashes to the HeavyHitters
             object (hence the different signature on the function!)
 
@@ -619,7 +636,7 @@ class HeavyHitters(CountMinSketch):
             self.__smallest = self.__top_x[new_min]
         return res
 
-    def remove_alt(self, hashes, num_els=1):
+    def remove_alt(self, hashes: HashResultsT, num_els: int = 1):
         """ Remove element based on hashes provided; not supported in
             heavy hitters
 
@@ -635,14 +652,14 @@ class HeavyHitters(CountMinSketch):
         )
         raise NotSupportedError(msg)
 
-    def clear(self):
-        """ Clear out the heavy hitters! """
+    def clear(self) -> None:
+        """Clear out the heavy hitters!"""
         super(HeavyHitters, self).clear()
         self.__top_x = dict()
         self.__top_x_size = 0
         self.__smallest = 0
 
-    def join(self, second):
+    def join(self, second: "HeavyHitters"):  # type: ignore
         """ Join is not supported by HeavyHitters
 
             Raises:
@@ -653,46 +670,46 @@ class HeavyHitters(CountMinSketch):
 
 
 class StreamThreshold(CountMinSketch):
-    """ keep track of those elements over a certain threshold """
+    """keep track of those elements over a certain threshold"""
 
     __slots__ = ["__threshold", "__meets_threshold"]
 
     def __init__(
         self,
-        threshold=100,
-        width=None,
-        depth=None,
-        confidence=None,
-        error_rate=None,
-        filepath=None,
-        hash_function=None,
-    ):
+        threshold: int = 100,
+        width: typing.Optional[int] = None,
+        depth: typing.Optional[int] = None,
+        confidence: typing.Optional[float] = None,
+        error_rate: typing.Optional[float] = None,
+        filepath: typing.Optional[str] = None,
+        hash_function: typing.Optional[HashFuncT] = None,
+    ) -> None:
         super(StreamThreshold, self).__init__(width, depth, confidence, error_rate, filepath, hash_function)
         self.__threshold = threshold
-        self.__meets_threshold = dict()
+        self.__meets_threshold = dict()  # type: ignore
 
-    def __str__(self):
-        """ stream threshold string rep """
+    def __str__(self) -> str:
+        """stream threshold string rep"""
         msg = super(StreamThreshold, self).__str__()
-        tmp = "Stream Threshold {0}\n" "\tThreshold: {1}\n" "\tNumber Meeting Threshold: {2}"
+        tmp = "Stream Threshold {0}\n\tThreshold: {1}\n\tNumber Meeting Threshold: {2}"
         return tmp.format(msg, self.threshold, len(self.__meets_threshold))
 
     @property
-    def meets_threshold(self):
-        """ dict: Those keys that meet the required threshold (with value) """
+    def meets_threshold(self) -> typing.Dict[str, int]:
+        """dict: Those keys that meet the required threshold (with value)"""
         return self.__meets_threshold
 
     @property
-    def threshold(self):
-        """ int: The threshold at which a key is tracked """
+    def threshold(self) -> int:
+        """int: The threshold at which a key is tracked"""
         return self.__threshold
 
-    def clear(self):
-        """ Clear out the stream threshold! """
+    def clear(self) -> None:
+        """Clear out the stream threshold!"""
         super(StreamThreshold, self).clear()
         self.__meets_threshold = dict()
 
-    def add(self, key, num_els=1):
+    def add(self, key: str, num_els: int = 1) -> int:  # type: ignore
         """Add the element for key into the data structure
 
         Args:
@@ -705,7 +722,7 @@ class StreamThreshold(CountMinSketch):
         hashes = self.hashes(key)
         return self.add_alt(key, hashes, num_els)
 
-    def add_alt(self, key, hashes, num_els=1):
+    def add_alt(self, key: str, hashes: HashResultsT, num_els: int = 1) -> int:  # type: ignore
         """ Add the element for key into the data structure
 
             Args:
@@ -724,7 +741,7 @@ class StreamThreshold(CountMinSketch):
             self.__meets_threshold[key] = res
         return res
 
-    def remove(self, key, num_els=1):
+    def remove(self, key: str, num_els: int = 1) -> int:  # type: ignore
         """ Remove element 'key' from the count-min sketch
 
             Args:
@@ -738,7 +755,7 @@ class StreamThreshold(CountMinSketch):
         hashes = self.hashes(key)
         return self.remove_alt(key, hashes, num_els)
 
-    def remove_alt(self, key, hashes, num_els=1):
+    def remove_alt(self, key: str, hashes: HashResultsT, num_els: int = 1) -> int:  # type: ignore
         """ Remove an element by using the hash representation
 
             Args:
@@ -759,7 +776,7 @@ class StreamThreshold(CountMinSketch):
             self.__meets_threshold[key] = res
         return res
 
-    def join(self, second):
+    def join(self, second: "StreamThreshold"):  # type: ignore
         """ Join is not supported by StreamThreshold
 
             Raises:
